/* graph.h */
/*
    Copyright Vladimir Kolmogorov (vnk@ist.ac.at), Yuri Boykov (yuri@csd.uwo.ca) 

    This file is part of MAXFLOW.

    MAXFLOW is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    MAXFLOW is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with MAXFLOW.  If not, see <http://www.gnu.org/licenses/>.

========================

	version 3.04
	+ modifications for GomoryHu (Vladimir Kolmogorov, 2022)

	This software library implements the maxflow algorithm
	described in

		"An Experimental Comparison of Min-Cut/Max-Flow Algorithms for Energy Minimization in Vision."
		Yuri Boykov and Vladimir Kolmogorov.
		In IEEE Transactions on Pattern Analysis and Machine Intelligence (PAMI), 
		September 2004

	This algorithm was developed by Yuri Boykov and Vladimir Kolmogorov
	at Siemens Corporate Research. To make it available for public use,
	it was later reimplemented by Vladimir Kolmogorov based on open publications.

	If you use this software for research purposes, you should cite
	the aforementioned paper in any resulting publication.

	----------------------------------------------------------------------

	REUSING TREES:

	Starting with version 3.0, there is a also an option of reusing search
	trees from one maxflow computation to the next, as described in

		"Efficiently Solving Dynamic Markov Random Fields Using Graph Cuts."
		Pushmeet Kohli and Philip H.S. Torr
		International Conference on Computer Vision (ICCV), 2005

	If you use this option, you should cite
	the aforementioned paper in any resulting publication.
*/
	


/*
	For description, license, example usage see README.TXT.
*/

#ifndef __GRAPH_H__
#define __GRAPH_H__


#include <string.h>
#include <iostream>
#include <assert.h>
#include "block.h"
#include "tmp_graph.h"

#define UPDATE_EXCESS_FIRST(g, i, a, b) (g)->add_tweights(i, a, b)
#define UPDATE_EXCESS_INC(g, i, a, b) { (g)->add_tweights(i, a, b); (g)->mark_node(i); }
#define MAXFLOW_SINGLE(g, flow) flow = (g)->maxflow()
#define MAXFLOW_FIRST(g) (g)->maxflow()
#define MAXFLOW_INC(g) (g)->maxflow(true)
#define _MAXFLOW_SINGLE(g, flow) flow = (g)->maxflow()
#define _MAXFLOW_FIRST(g) (g)->maxflow()
#define _MAXFLOW_INC(g) (g)->maxflow(true)

#ifdef GRAPH_NAMESPACE
namespace GRAPH_NAMESPACE {
#endif

// TS: timestamp showing when DIST was computed
// DIST: distance to the terminal
// flag: shows in which list the node is


#ifdef NODE_FIELDS
#undef NODE_FIELDS
#endif
#define NODE_FIELDS \
		node		*prev;\
		node		*next;\
		arc			*parent;\
		long		TS;\
		int			DIST;\
		unsigned int	is_sink : 1;\
		unsigned int flag : 2;\
		unsigned int L : 4;\
		unsigned int	is_in_changed_list : 1;

#ifndef GRAPH_BASE_FILE
#define GRAPH_BASE_FILE "graph_base_forward_star.h"
#endif
#include GRAPH_BASE_FILE

// NOTE: in UNIX you need to use -DNDEBUG preprocessor option to supress assert's!!!

#define ONE_SIDED_ACTIVE


	

#define TERMINAL ( (arc *) 1 )		/* to terminal */
#define ORPHAN   ( (arc *) 2 )		/* orphan */

// captype: type of edge capacities (excluding t-links)
// tcaptype: type of t-links (edges between nodes and terminals)
// flowtype: type of total flow
//
// Current instantiations are in instances.inc
class Graph : public GraphBase
{
public:
	typedef enum
	{
		SOURCE	= 0,
		SINK	= 1
	} termtype; // terminals 
	typedef int node_id;

	/////////////////////////////////////////////////////////////////////////
	//                     BASIC INTERFACE FUNCTIONS                       //
	//              (should be enough for most applications)               //
	/////////////////////////////////////////////////////////////////////////

	// Constructor. 
	// The first argument gives an estimate of the maximum number of nodes that can be added
	// to the graph, and the second argument is an estimate of the maximum number of edges.
	// The last (optional) argument is the pointer to the function which will be called 
	// if an error occurs; an error message is passed to this function. 
	// If this argument is omitted, exit(1) will be called.
	//
	// IMPORTANT: It is possible to add more nodes to the graph than node_num_max 
	// (and node_num_max can be zero). However, if the count is exceeded, then 
	// the internal memory is reallocated (increased by 50%) which is expensive. 
	// Also, temporarily the amount of allocated memory would be more than twice than needed.
	// Similarly for edges.
	// If you wish to avoid this overhead, you can download version 2.2, where nodes and edges are stored in blocks.
	//
	// If buf != NULL, it is used for allocating. # nodes and edges cannot be exceeded then.

#ifdef KEEP_STLISTS
	#define MAXFLOW_RBUF_NUM STLISTS_RBUF_NUM
#else
	#define MAXFLOW_RBUF_NUM 1  // to allow expression ReusableBuffer rbuf[MAXFLOW_RBUF_NUM]
#endif

	// rbuf must be of size MAXFLOW_RBUF_NUM. common_rbuf can be used outside, rbuf[i] cannot.
	Graph(int node_num, int edge_num_max, Buffer* buf, ReusableBuffer* common_rbuf, ReusableBuffer* rbuf)
		: GraphBase(node_num, edge_num_max, buf, common_rbuf, rbuf)
	{
		maxflow_iteration = 0;
		flow = 0;
		operation_count = 0;
	}

	// Destructor
	~Graph()
	{
	}

	// Adds a bidirectional edge between 'i' and 'j' with the weights 'cap' and 'rev_cap'.
	// IMPORTANT: see note about the constructor 
	//void add_edge(node_id i, node_id j, captype cap, captype rev_cap);

	// Adds new edges 'SOURCE->i' and 'i->SINK' with corresponding weights.
	// Can be called multiple times for each node.
	// Weights can be negative.
	// NOTE: the number of such edges is not counted in edge_num_max.
	//       No internal memory is allocated by this call.
	void add_tweights(node_id i, tcaptype cap_source, tcaptype cap_sink);


	// Computes the maxflow. Can be called several times.
	// FOR DESCRIPTION OF reuse_trees, SEE mark_node().
	// FOR DESCRIPTION OF changed_list, SEE remove_from_changed_list().
	flowtype maxflow(bool reuse_trees = false, Block<node_id>* changed_list = NULL);

	// After the maxflow is computed, this function returns to which
	// segment the node 'i' belongs (Graph::SOURCE or Graph::SINK).
	//
	// Occasionally there may be several minimum cuts. If a node can be assigned
	// to both the source and the sink, then default_segm is returned.
#ifdef ONE_SIDED_ACTIVE
	termtype what_segment(node_id i);
#else
	termtype what_segment(node_id i, termtype default_segm = SOURCE);
#endif
//	void set_segment_flags()
//	{
//		int i;
//		FOR_ALL_NODES(i) nodes[i].segment = what_segment(i);
//	}

	//////////////////////////////////////////////
	//       ADVANCED INTERFACE FUNCTIONS       //
	//      (provide access to the graph)       //
	//////////////////////////////////////////////

//protected:
	typedef GraphBase::node node;
	typedef GraphBase::arc arc;

public:

	///////////////////////////////////////////////////
	// 3. Functions for reading residual capacities. //
	///////////////////////////////////////////////////

	// returns residual capacity of SOURCE->i minus residual capacity of i->SINK
	tcaptype get_trcap(node_id i); 

	/////////////////////////////////////////////////////////////////
	// 4. Functions for setting residual capacities.               //
	//    NOTE: If these functions are used, the value of the flow //
	//    returned by maxflow() will not be valid!                 //
	/////////////////////////////////////////////////////////////////

	void set_trcap(node_id i, tcaptype trcap); 

	////////////////////////////////////////////////////////////////////
	// 5. Functions related to reusing trees & list of changed nodes. //
	////////////////////////////////////////////////////////////////////

	// If flag reuse_trees is true while calling maxflow(), then search trees
	// are reused from previous maxflow computation. 
	// In this case before calling maxflow() the user must
	// specify which parts of the graph have changed by calling mark_node():
	//   add_tweights(i),set_trcap(i)    => call mark_node(i)
	//   add_edge(i,j),set_rcap(a)       => call mark_node(i); mark_node(j)
	//
	// This option makes sense only if a small part of the graph is changed.
	// The initialization procedure goes only through marked nodes then.
	// 
	// mark_node(i) can either be called before or after graph modification.
	// Can be called more than once per node, but calls after the first one
	// do not have any effect.
	// 
	// NOTE: 
	//   - This option cannot be used in the first call to maxflow().
	//   - It is not necessary to call mark_node() if the change is ``not essential'',
	//     i.e. sign(trcap) is preserved for a node and zero/nonzero status is preserved for an arc.
	//   - To check that you marked all necessary nodes, you can call maxflow(false) after calling maxflow(true).
	//     If everything is correct, the two calls must return the same value of flow. (Useful for debugging).
	void mark_node(node_id i);

	// If changed_list is not NULL while calling maxflow(), then the algorithm
	// keeps a list of nodes which could potentially have changed their segmentation label.
	// Nodes which are not in the list are guaranteed to keep their old segmentation label (SOURCE or SINK).
	// Example usage:
	//
	//		typedef Graph<int,int,int> G;
	//		G* g = new Graph(nodeNum, edgeNum);
	//		Block<G::node_id>* changed_list = new Block<G::node_id>(128);
	//
	//		... // add nodes and edges
	//
	//		g->maxflow(); // first call should be without arguments
	//		for (int iter=0; iter<10; iter++)
	//		{
	//			... // change graph, call mark_node() accordingly
	//
	//			g->maxflow(true, changed_list);
	//			G::node_id* ptr;
	//			for (ptr=changed_list->ScanFirst(); ptr; ptr=changed_list->ScanNext())
	//			{
	//				G::node_id i = *ptr; assert(i>=0 && i<nodeNum);
	//				g->remove_from_changed_list(i);
	//				// do something with node i...
	//				if (g->what_segment(i) == G::SOURCE) { ... }
	//			}
	//			changed_list->Reset();
	//		}
	//		delete changed_list;
	//		
	// NOTE:
	//  - If changed_list option is used, then reuse_trees must be used as well.
	//  - In the example above, the user may omit calls g->remove_from_changed_list(i) and changed_list->Reset() in a given iteration.
	//    Then during the next call to maxflow(true, &changed_list) new nodes will be added to changed_list.
	//  - If the next call to maxflow() does not use option reuse_trees, then calling remove_from_changed_list()
	//    is not necessary. ("changed_list->Reset()" or "delete changed_list" should still be called, though).
	void remove_from_changed_list(node_id i) 
	{ 
		assert(i>=0 && i<node_num && nodes[i].is_in_changed_list); 
		nodes[i].is_in_changed_list = 0;
	}

#ifdef GRAPH_CONTRACTION
	void contract(int num, node_id* list);
	void contract_to_source();
#endif

/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
	
protected:
	friend class GomoryHu;
	// internal variables and functions

#define FLAG_NONE 0
#define FLAG_ACTIVE 1
#define FLAG_ORPHAN 2
#define FLAG_MARKED 3

	flowtype			flow;		// total flow

	// reusing trees & list of changed pixels
	int					maxflow_iteration; // counter
	Block<node_id>		*changed_list;

	/////////////////////////////////////////////////////////////////////////

	struct Lists
	{
#define ACTIVE_BUCKET_NUM 2

#define QUEUE_ACTIVE_SOURCE     0
#define QUEUE_ACTIVE_SINK       ACTIVE_BUCKET_NUM
#define QUEUE_ORPHAN            (2*ACTIVE_BUCKET_NUM)
#define QUEUE_MARKED            (2*ACTIVE_BUCKET_NUM+1)

#define QUEUE_NUM (2*ACTIVE_BUCKET_NUM + 2)

		void reset()
		{
			int L;
			for (L=0; L<QUEUE_NUM; L++) queue_first[L] = queue_last[L] = NULL;
			active_num[0] = active_num[1] = 0;
			active_L_source = active_L_sink = 0;
		}

		node* next_active()
		{
#ifdef ONE_SIDED_ACTIVE
			if (active_num[0] == 0) { default_segm = SINK; return NULL; }
			if (active_num[1] == 0) { default_segm = SOURCE; return NULL; }
#else
			if (active_num[0] == 0) return next_active_sink();
			if (active_num[1] == 0) return next_active_source();
#endif
			return (active_num[0] < active_num[1]) ? next_active_source() : next_active_sink();
		}

		node* next_orphan()
		{
			node* i = queue_first[QUEUE_ORPHAN];
			if (i) remove_from_list(i);
			return i;
		}

		node* next_marked()
		{
			node* i = queue_first[QUEUE_MARKED];
			if (i) remove_from_list(i);
			return i;
		}

		void add_to_active_front(node* i)
		{
			if (i->flag) remove_from_list(i);
			i->flag = FLAG_ACTIVE;
			if (i->is_sink) { add_to_list_front(i, active_L_sink   + QUEUE_ACTIVE_SINK);   active_num[1] ++; }
			else            { add_to_list_front(i, active_L_source + QUEUE_ACTIVE_SOURCE); active_num[0] ++; }
		}
		void add_to_active_rear(node* i)
		{
			if (i->flag) remove_from_list(i);
			i->flag = FLAG_ACTIVE;
			if (i->is_sink) { add_to_list_rear(i, 1-active_L_sink   + QUEUE_ACTIVE_SINK);   active_num[1] ++; }
			else            { add_to_list_rear(i, 1-active_L_source + QUEUE_ACTIVE_SOURCE); active_num[0] ++; }
		}
		void add_to_orphan_front(node* i)
		{
			if (i->flag) remove_from_list(i);
			i->flag = FLAG_ORPHAN;
			i->parent = ORPHAN;
			add_to_list_front(i, QUEUE_ORPHAN);
		}
		void add_to_orphan_rear(node* i) 
		{
			if (i->flag) remove_from_list(i);
			i->flag = FLAG_ORPHAN;
			i->parent = ORPHAN;
			add_to_list_rear(i,  QUEUE_ORPHAN);
		}

		void add_to_marked(node* i)
		{
			if (i->flag) remove_from_list(i);
			i->flag = FLAG_MARKED;
			add_to_list_rear(i, QUEUE_MARKED);
		}

		void remove_from_list(node* i)
		{
			assert(i->flag != 0);
			if (i->flag == FLAG_ACTIVE) active_num[i->is_sink] --;
			if (i->prev) i->prev->next = i->next;
			else         queue_first[i->L] = i->next;
			if (i->next) i->next->prev = i->prev;
			else         queue_last[i->L] = i->prev;
			i->flag = FLAG_NONE;
		}

		struct OrphanListData
		{
			node* first;
			node* last;
		};
		void save_orphan_list(OrphanListData& D) { D.first = queue_first[QUEUE_ORPHAN]; D.last = queue_last[QUEUE_ORPHAN]; D.first = D.last = NULL; } 
		void restore_orphan_list(OrphanListData& D) { queue_first[QUEUE_ORPHAN] = D.first; queue_last[QUEUE_ORPHAN] = D.last; }

#ifdef ONE_SIDED_ACTIVE
		termtype default_segm;
#endif
	private:

		node* next_active_source()
		{
			if (!queue_first[QUEUE_ACTIVE_SOURCE+active_L_source])
			{
				active_L_source = 1 - active_L_source;
			}
			node* i = queue_first[QUEUE_ACTIVE_SOURCE+active_L_source];
			if (i)
			{
				assert(i->parent && !i->is_sink);
				remove_from_list(i);
			}
			return i;
		}
		node* next_active_sink()
		{
			if (!queue_first[QUEUE_ACTIVE_SINK+active_L_sink])
			{
				active_L_sink = 1 - active_L_sink;
			}
			node* i = queue_first[QUEUE_ACTIVE_SINK+active_L_sink];
			if (i)
			{
				assert(i->parent && i->is_sink);
				remove_from_list(i);
			}
			return i;
		}

		void add_to_list_front(node* i, int L)
		{
			i->L = L;
			i->next = queue_first[L];
			if (queue_first[L]) queue_first[L] -> prev = i;
			else                queue_last[L]          = i;
			queue_first[L] = i;
			i -> prev = NULL;
		}
		void add_to_list_rear(node* i, int L)
		{
			i->L = L;
			i->prev = queue_last[L];
			if (queue_last[L]) queue_last[L] -> next = i;
			else               queue_first[L]        = i;
			queue_last[L] = i;
			i -> next = NULL;
		}


		node *queue_first[QUEUE_NUM], *queue_last[QUEUE_NUM];
		int active_L_source, active_L_sink;
		int active_num[2];

	} L;

	int					operation_count;
	long				TIME;								// monotonically increasing global counter
															
	/////////////////////////////////////////////////////////////////////////

	void reallocate_nodes(int num); // num is the number of new nodes
	void reallocate_arcs();

	// functions for processing orphans list
	void add_to_changed_list(node* i);

	void maxflow_init();             // called if reuse_trees == false
	void maxflow_reuse_trees_init(); // called if reuse_trees == true
	void augment(arc *middle_arc);
	void process_source_orphan(node *i);
	void process_sink_orphan(node *i);

	void test_consistency(node* current_node=NULL); // debug function
};











///////////////////////////////////////
// Implementation - inline functions //
///////////////////////////////////////



 
 
inline void Graph::add_tweights(node_id _i, tcaptype cap_source, tcaptype cap_sink)
{
	assert(_i >= 0 && _i < node_num);
	node* i = nodes + _i;

	tcaptype delta = i->excess;
	if (delta > 0) cap_source += delta;
	else           cap_sink   -= delta;
	flow += (cap_source < cap_sink) ? cap_source : cap_sink;
	SET_EXCESS(i, cap_source - cap_sink);
}

inline Graph::captype Graph::get_trcap(node_id i)
{
	assert(i>=0 && i<node_num);
	return nodes[i].excess;
}


inline void Graph::set_trcap(node_id i, tcaptype trcap)
{
	assert(i>=0 && i<node_num); 
	SET_EXCESS(nodes+i, trcap);
}

 
inline Graph::termtype Graph::what_segment(node_id i
#ifdef ONE_SIDED_ACTIVE
	)
#else
	, termtype default_segm)
#endif
{
	if (nodes[i].parent)
	{
		return (nodes[i].is_sink) ? SINK : SOURCE;
	}
	else
	{
#ifdef ONE_SIDED_ACTIVE
		return L.default_segm;
#else
		return default_segm;
#endif
	}
}

 
inline void Graph::mark_node(node_id _i)
{
#ifdef GRAPH_CONTRACTION
	if (nodes[_i].is_contracted) return;
#endif
	L.add_to_marked(nodes + _i);
}

#ifdef GRAPH_CONTRACTION
inline void Graph::contract(int num, node_id* list)
{
	int k;
	for (k=1; k<num; k++)
	{
		node* i = nodes + list[k];
		if (i->flag) L.remove_from_list(i);
	}

	GraphBase::contract(num, list);

	arc* a;
	node* r = nodes + list[0];
	r->is_contracted = 1;
	FOR_ALL_ARCS(r, a)
	{
		node* i = a->head;
		if (i->parent && i->parent != TERMINAL && i->parent->head->is_contracted) i->parent = REV_ARC(a);
		if (i->flag) L.remove_from_list(i);
		mark_node((int)(i-nodes));
	}
	r->is_contracted = 0;
	mark_node(list[0]);
}

inline void Graph::contract_to_source()
{
	GraphBase::contract_to_source();
}

#endif

#ifdef GRAPH_NAMESPACE
}
#endif

#endif
