#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <vector>
#include <algorithm>
#include "gomory_hu.h"
#include "ordered_cuts.h"

//#define OC_DEBUG

extern double GH_total_node_num;
extern double GH_total_edge_num;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////


struct GomoryHu
{
	GomoryHu(TmpGraph<GH_REAL>* input, GomoryHuOptions* options);
	~GomoryHu();

	void UnmarkNode(int i) { nodes[i].is_fixed = 1; n_nonfixed --; }
	void Solve(int* parents, GH_REAL* cut_costs);

	void PrintGraph()
	{
		int i, dir;
		Edge* e;
		for (i=0; i<n; i++)
		{
			if (nodes[i].is_fixed) printf(". "); else printf("* ");
			for (dir=0; dir<2; dir++)
			for (e=nodes[i].first[dir]; e; e=e->next[dir])
			{
				printf("[%d %d|%d] ", e->i[dir], e->i[1-dir], (int)e->w);
			}
		}
		printf("\n");
	}

private:

	struct Edge;
	struct Node
	{
		Edge*	first[2];
		GH_REAL	upper_bound;
		int		index;
		int		rank; // 0: not in OClist. >0: in OClist. 
		unsigned int is_fixed : 1;
		unsigned int is_component_root : 1;
		unsigned int is_contracted : 1;
		unsigned int tmp_flag : 1; // must be kept 0
	};
	struct Edge
	{
		int		i[2];
		Edge*	next[2];
		GH_REAL	w;
	};
	Node* nodes;
	Edge* edges;
	int n_orig, n, n_nonfixed, m, m_max;
	int* mapping_orig; // of size n
	int* parents_orig; // of size n_orig
	GH_REAL* cut_costs_orig; // of size n_orig

	int s, L; // L <= n-1
	int* OClist; // of size L
	OrderedCuts::Node* OCnodes; // of size n

	int component_first_node;

	GomoryHu(GomoryHu* parent);
	void Init(int n, int m);
	void AddEdge(int i, int j, GH_REAL w)
	{
		if (m >= m_max) { printf("Too many edges\n"); exit(1); }
		Edge* e = &edges[m ++];
		e->i[0] = i; e->i[1] = j;
		e->next[0] = nodes[i].first[0];
		e->next[1] = nodes[j].first[1];
		nodes[i].first[0] = nodes[j].first[1] = e;
		e->w = w;
	}

	void ComputeOCTree();
	void PrepareComponents();
	int ProcessComponent(int num, int* list, GH_REAL cost, int s_new=-1);
	void Compact();

	ReusableBuffer* common_rbuf;
	ReusableBuffer local_rbuf;
	OrderedCuts* C;
	GH_REAL* buf_REAL; // of size n
	int*	buf1_int; // of size n
	int*	buf2_int; // of size n, used in Solve()

	GomoryHuOptions* options;
	int depth;
	GomoryHu* parent;
	GomoryHu* child;
	GH_REAL INFTY;
};




////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////

GomoryHu::GomoryHu(TmpGraph<GH_REAL>* input, GomoryHuOptions* _options) : parent(NULL), child(NULL), depth(0), options(_options)
{
	int i, j, e;
	Init(input->n, input->m);
	for (e=0; e<input->m; e++)
	{
		i = input->edges[2*e];
		j = input->edges[2*e+1];
		GH_REAL w = input->weights[e];
		nodes[i].upper_bound += w;
		nodes[j].upper_bound += w;
		AddEdge(i, j, w);
	}
	if (n > 0)
	{
		INFTY = nodes[0].upper_bound;
		s = 0;
		for (i=1; i<input->n; i++) { if (INFTY < nodes[i].upper_bound) { s = i; INFTY = nodes[i].upper_bound; } }
		INFTY += 1;
	}
	common_rbuf = new ReusableBuffer;
	C = new OrderedCuts(INFTY, options, common_rbuf);
}


GomoryHu::GomoryHu(GomoryHu* _parent) : depth(_parent->depth+1), parent(_parent), child(NULL)
{
	common_rbuf = parent->common_rbuf;
	C = parent->C;
	INFTY = parent->INFTY;
	options = parent->options;
}

GomoryHu::~GomoryHu()
{
	if (child)
	{
		delete child;
	}
	if (!parent)
	{
		delete C;
		delete common_rbuf;
	}
}


void GomoryHu::Init(int _n, int _m)
{
	int i;
	n_orig = n = n_nonfixed = _n;
	L = 0;
	m = 0;
	m_max = _m;
	nodes = (Node*) local_rbuf.Alloc(n*sizeof(Node) + m_max*sizeof(Edge) + 2*n*sizeof(int) + n*sizeof(OrderedCuts::Node) + n*sizeof(GH_REAL) + 2*n*sizeof(int));
	edges = (Edge*) (nodes + n);
	mapping_orig = (int*) (edges + m_max);
	OClist = (int*) (mapping_orig + n);
	OCnodes = (OrderedCuts::Node*) (OClist + n);

	buf_REAL = (GH_REAL*) (OCnodes + n);
	buf1_int = (int*) (buf_REAL + n);
	buf2_int = buf1_int + n;

	memset(nodes, 0, n*sizeof(Node));
	for (i=0; i<n; i++) mapping_orig[i] = i;

	component_first_node = -1;
}


void GomoryHu::ComputeOCTree()
{
	if (n_nonfixed < 1) return;
	int i, j;
	Edge* e;

	{
		std::vector<int> V(n);
		for (i=L=0; i<n; i++)
		{
			//if (i != s && !nodes[i].is_fixed && rand()%2) V[L ++] = i;
			if (i != s && !nodes[i].is_fixed) V[L ++] = i;
		}
		sort( V.begin(),V.begin()+L, [&](int i,int j)
			{return (nodes[i].upper_bound > nodes[j].upper_bound) || (nodes[i].upper_bound == nodes[j].upper_bound && nodes[i].rank > nodes[j].rank);}
		);

		for (i=0; i<n; i++) nodes[i].rank = 0;
		for (i=0; i<L; i++)
		{
			nodes[V[i]].rank = n-i;
			OClist[i] = V[i];
		}
	}

	GH_REAL* unaries = buf_REAL;
	memset(unaries, 0, n*sizeof(GH_REAL));
	int* OClist_copy = buf1_int;
	memcpy(OClist_copy, OClist, L*sizeof(int));

	TmpGraph<GH_REAL> input(n, m);

	for (i=0; i<n; i++)
	{
		for (e=nodes[i].first[0]; e; e=e->next[0])
		{
			if (e->w <= 0) continue;
			j = e->i[1];
			if (i == s) unaries[j] += e->w;
			else if (j == s) unaries[i] += e->w;
			else input.AddEdge(i, j, e->w);
		}
	}
#ifdef OC_DEBUG
	printf("calling OrderedCuts:\n");
	input.Print();
	printf("unaries: ");
	for (i=0; i<n; i++) printf("%d ", unaries[i]);
	printf("\nlist:");
	for (i=0; i<L; i++) printf("%d ", OClist_copy[i]);
	printf("\n");
#endif
	if (options->parametric_ordered_cuts)
	{
		C->SolveParametric(&input, unaries, L, OClist_copy, OCnodes);
	}
	else
	{
		C->Solve(&input, unaries, L, OClist_copy, OCnodes);
	}
#ifdef OC_DEBUG
	printf("result: ");
	for (i=0; i<n; i++) printf("%d:%d [%d] ", i, OCnodes[i].parent, OCnodes[i].cost);
	printf("\n");
#endif
	for (i=0; i<n; i++)
	{
		if (OCnodes[i].parent < 0 && i != s) OCnodes[i].parent = s;
	}
}

void GomoryHu::PrepareComponents()
{
	int k, i, num = 1;
	GH_REAL* best_cost = (GH_REAL*) common_rbuf->Alloc(2*n*sizeof(GH_REAL) + 2*n*sizeof(int));
	GH_REAL* upper_bounds = best_cost + n;
	int* list_first = (int*)(upper_bounds + n);
	int* list_last = list_first + n;

	nodes[s].is_component_root = 1;
	nodes[s].index = 0;
	upper_bounds[s] = best_cost[s] = INFTY;
	list_first[0] = list_last[0] = s;
	for (k=0; k<L; k++)
	{
		i = OClist[k];
		int p = OCnodes[i].parent;
		GH_REAL w = OCnodes[i].cost;
		upper_bounds[i] = (upper_bounds[p] < w) ? upper_bounds[p] : w;
		best_cost[i] = best_cost[p];
		if (w <= best_cost[p])
		{
			list_first[num] = list_last[num] = i;
			nodes[i].is_component_root = 1;
			nodes[i].index = num ++;
			best_cost[i] = best_cost[p] = w;
#ifdef OC_DEBUG
			if (w > nodes[i].upper_bound)
			{
				printf("wrong upper bound!\n");
				exit(1);
			}
#endif
		}
		else
		{
			nodes[i].index = nodes[p].index;
			if (!nodes[p].is_component_root && OCnodes[i].cost > OCnodes[p].cost) OCnodes[i].cost = OCnodes[p].cost;
		}
	}
	for (i=0; i<n; i++)
	{
		if (i == s) continue;
		if (nodes[i].rank > 0)
		{
			if (nodes[i].upper_bound > upper_bounds[i]) nodes[i].upper_bound = upper_bounds[i];
		}
		else
		{
			int p = OCnodes[i].parent;
			nodes[i].index = nodes[p].index;
			if (nodes[i].upper_bound > upper_bounds[p]) nodes[i].upper_bound = upper_bounds[p];
		}
	}
	for (i=0; i<n; i++)
	{
		if (!nodes[i].is_component_root)
		{
			k = nodes[i].index;
			nodes[list_last[k]].index = i;
			list_last[k] = i;
		}
	}

	component_first_node = list_first[num-1];
	for (k=num-1; k>0; k--)
	{
		nodes[list_last[k]].index = list_first[k-1];
	}
	nodes[list_last[0]].index = n;
}

void ChangeRoot(int n, int r_old, int r_new, int* parents, GH_REAL* costs)
{
	int i = r_new;
	int j = parents[i];
	GH_REAL w = costs[i];
	while (i != r_old)
	{
		int k = parents[j];
		parents[j] = i;
		i = j;
		j = k;
		GH_REAL x = w; w = costs[i]; costs[i] = x;
	}
	parents[r_new] = -1;
	costs[r_new] = 0;
}

int GomoryHu::ProcessComponent(int num, int* list, GH_REAL cost, int s_new)
{
	int i, j, k, dir;
	Edge* e;

	int* index_old = NULL;
	if (s_new >= 0)
	{
		index_old = (int*) common_rbuf->Alloc(n*sizeof(int));
		for (i=0; i<n; i++) index_old[i] = nodes[i].index;
	}

	int m_new = 0;
	for (k=0; k<num; k++)
	{
		i = list[k];
		nodes[i].index = k+1;
		for (dir=0; dir<2; dir++)
		for (e=nodes[i].first[dir]; e; e=e->next[dir])
		{
			if (dir == 0 || !nodes[e->i[1-dir]].is_contracted) m_new ++;
		}
	}
	if (m_new > m) m_new = m;

	if (!child)
	{
		child = new GomoryHu(this);
	}
	child->Init(num+1, m_new);
	child->s = 1;

	if (s_new >= 0) // source component
	{
		for (k=0; k<num; k++)
		{
			i = list[k];
			child->nodes[k+1].is_component_root = nodes[i].is_component_root;
			nodes[i].is_component_root = 0;
			child->nodes[k+1].index = (index_old[i] < n) ? nodes[index_old[i]].index : (num+1);

			child->OCnodes[k+1].parent = (OCnodes[i].parent >= 0) ? nodes[OCnodes[i].parent].index : -1;
			child->OCnodes[k+1].cost = OCnodes[i].cost;
		}
		child->component_first_node = nodes[component_first_node].index;
		int p = nodes[OCnodes[s_new].parent].index;
		child->nodes[0].index = child->nodes[p].index;
		child->nodes[p].index = 0;

		child->OCnodes[0].parent = p;
		child->OCnodes[0].cost = cost;
	}

	Edge** e_contracted = (Edge**) common_rbuf->Alloc(n*sizeof(Edge*));
	Edge* e_first[2] = { NULL, NULL };
	Edge* e_last[2] = { NULL, NULL };

	child->UnmarkNode(0);
	for (k=0; k<num; k++)
	{
		i = list[k];
		if (nodes[i].is_fixed) child->UnmarkNode(k+1);
		else
		{
			nodes[i].is_fixed = 1;
			n_nonfixed --;
			child->nodes[k+1].upper_bound = (s_new >= 0) ? nodes[i].upper_bound : OCnodes[i].cost;
			child->nodes[k+1].rank = nodes[i].rank;
		}
		GH_REAL w_sum = 0;
		for (dir=0; dir<2; dir++)
		for (e=nodes[i].first[dir]; e; e=e->next[dir])
		{
			if (e->w <= 0) continue;
			j = e->i[1-dir];
			if (nodes[j].is_contracted)
			{
				if (dir == 0) child->AddEdge(k+1, nodes[j].index, e->w);
				continue;
			}
			w_sum += e->w;
			if (!nodes[j].tmp_flag)
			{
				nodes[j].tmp_flag = 1;
				e_contracted[j] = e;
				if (e_last[dir]) e_last[dir]->next[dir] = e;
				else             e_first[dir] = e;
				e_last[dir] = e;
			}
			else
			{
				e_contracted[j]->w += e->w;
				e->w = 0;
			}
		}
		if (w_sum > 0)
		{
			child->AddEdge(0, k+1, w_sum);
		}
	}

	int* parents = buf1_int;
	GH_REAL* cut_costs = buf_REAL;

	child->Solve(parents, cut_costs);

	int v = parents[0];
	if (v <= 0) { printf("Error #1 in ProcessComponent()\n"); exit(1); }
	v = list[v-1];

	if (parents[child->s] >= 0) { printf("Error #2 in ProcessComponent()\n"); exit(1); }
	ChangeRoot(k+1, child->s, 0, parents, cut_costs);

	cut_costs_orig[mapping_orig[v]] = cost;
	for (k=0; k<num; k++)
	{
		i = list[k];
		int p = parents[k+1];
		if (p == 0)
		{
			if (i != v) { printf("Error #3 in ProcessComponent()\n"); exit(1); }
			continue;
		}
		parents_orig[mapping_orig[i]] = mapping_orig[list[p-1]];
		cut_costs_orig[mapping_orig[i]] = cut_costs[k+1];
	}

	nodes[v].first[0] = e_first[0];
	nodes[v].first[1] = e_first[1];
	if (e_last[0]) e_last[0]->next[0] = NULL;
	if (e_last[1]) e_last[1]->next[1] = NULL;

	for (dir=0; dir<2; dir++)
	for (e=nodes[v].first[dir]; e; e=e->next[dir])
	{
		e->i[dir] = v;
		nodes[e->i[1-dir]].tmp_flag = 0;
	}

	nodes[v].is_contracted = 0;

	return v;
}

void GomoryHu::Compact()
{
	int i, dir;
	Edge* e;

	int* index = (int*) common_rbuf->Alloc(n*sizeof(int));

	int n_old = n;
	n = 0;
	for (i=0; i<n_old; i++)
	{
		if (!nodes[i].is_contracted)
		{
			index[i] = n;
			mapping_orig[n] = mapping_orig[i];
			nodes[n ++] = nodes[i];
		}
		else index[i] = -1;
	}

	m = 0;
	for (i=0; i<n; i++)
	{
		for (dir=0; dir<2; dir++)
		{
			Edge** e_prev_ptr = &nodes[i].first[dir];
			while ((e=*e_prev_ptr))
			{
				if (e->w > 0) { e_prev_ptr = &e->next[dir]; m ++; }
				else *e_prev_ptr = e->next[dir];
				e->i[dir] = index[e->i[dir]];
			}
		}
	}
	m /= 2;

	s = index[s];
}

void GomoryHu::Solve(int* _parents, GH_REAL* _cut_costs)
{
	parents_orig = _parents;
	cut_costs_orig = _cut_costs;

	int i;
	int* list = buf2_int;

	while (n_nonfixed > 1)
	{
		bool changed_source = false;
		if (options->verbose > depth+1)
		{
			for (int d=0; d<depth; d++) printf("    ");
			printf("(n n_nonfixed)=(%d %d) ... ", n, n_nonfixed);
		}
		GH_total_node_num += n;
		GH_total_edge_num += m;

		if (component_first_node < 0)
		{
			ComputeOCTree();
			PrepareComponents();
		}

		while (component_first_node != s)
		{
			int num = 0, num_nonfixed = 0;
			int r = component_first_node;
			while ( 1 )
			{
				list[num ++] = component_first_node;
				nodes[component_first_node].is_contracted = 1;
				component_first_node = nodes[component_first_node].index;
				if (!nodes[component_first_node].is_fixed) num_nonfixed ++;
				if (nodes[component_first_node].is_component_root) break;
			}
			if (num_nonfixed > n_nonfixed / 2)
			{
				for (i=0; i<num; i++) { nodes[list[i]].tmp_flag = 1; nodes[list[i]].is_contracted = 0; }
				num = 1;
				list[0] = s;
				nodes[s].is_contracted = 1;
				for (i=0; i<n; i++)
				{
					if (i == s || nodes[i].is_contracted) continue;
					if (nodes[i].tmp_flag)
					{
						nodes[i].tmp_flag = 0;
						nodes[i].upper_bound = OCnodes[i].cost;
					}
					else
					{
						list[num ++] = i;
						nodes[i].is_contracted = 1;
					}
				}
				ProcessComponent(num, list, OCnodes[r].cost, r);
				s = r;
				changed_source = true;
				break;
			}
			int v = ProcessComponent(num, list, OCnodes[r].cost);
			int p = OCnodes[r].parent;
			nodes[v].index = nodes[p].index;
			nodes[p].index = v;
			nodes[v].is_component_root = 0;
		}
		nodes[s].is_component_root = 0;
		component_first_node = -1;
		Compact();

		if (options->verbose > depth+1)
		{
			if (changed_source) printf("done (changed source)\n");
			else                printf("done\n");
		}
	}

	Edge* e;
	int dir;
	for (i=0; i<n; i++)
	{
		if (i == s) continue;
		parents_orig[mapping_orig[i]] = mapping_orig[s];
		GH_REAL cost = 0;
		for (dir=0; dir<2; dir++)
		for (e=nodes[i].first[dir]; e; e=e->next[dir])
		{
			cost += e->w;
		}
		cut_costs_orig[mapping_orig[i]] = cost;
	}
	s = mapping_orig[s];
	parents_orig[s] = -1;
	cut_costs_orig[s] = 0;
}





void _SolveGomoryHu(TmpGraph<GH_REAL>* input, int* parents, GH_REAL* cut_costs, GomoryHuOptions& options)
{
	GomoryHu S(input, &options);
	if (input->marks)
	{
		int i;
		for (i=0; i<input->n; i++)
		{
			if (!input->marks[i]) S.UnmarkNode(i);
		}
	}
//	S.PrintGraph();
	S.Solve(parents, cut_costs);

	if ( 0 )
	{
		int i;
		for (i=0; i<input->n; i++) printf("%d [%d] ", parents[i], cut_costs[i]);
		printf("\n");
	}
}

