#ifndef __GOMORY_HU_REDUCTIONS__
#define __GOMORY_HU_REDUCTIONS__

#include "tmp_graph.h"
#include "gomory_hu.h"

// usage:
// 
// GomoryHuReductions R(input, parents, cut_costs, int method);
// TmpGraph<GH_REAL>* subproblem;
// int* subproblem_parents;
// GH_REAL* subproblem_cut_costs;
// while ((subproblem=R.GetNextSubproblem(subproblem_parents, subproblem_cut_costs)))
// {
//   ... solve subproblem, write result to 'subproblem_parents' and 'subproblem_cut_costs'
// }
// ... now 'parents' and 'cut_costs' contain the final result

class GomoryHuReductions
{
public:
	GomoryHuReductions(TmpGraph<GH_REAL>* input, int* parents, GH_REAL* cut_costs, int method);
	~GomoryHuReductions();

	TmpGraph<GH_REAL>* GetNextSubproblem(int*& parents, GH_REAL*& cut_costs);

private:
	TmpGraph<GH_REAL>* input;
	int* parents;
	GH_REAL* cut_costs;
	int reduction_flags;

	TmpGraph<GH_REAL> I;
	int C, N, M;
};

GomoryHuReductions::GomoryHuReductions(TmpGraph<GH_REAL>* _input, int* _parents, GH_REAL* _cut_costs, int reduction_method)
	: input(_input), parents(_parents), cut_costs(_cut_costs), reduction_flags(reduction_method)
{
	if (reduction_flags & REDUCTION_CC)
	{
		if (reduction_flags & REDUCTION_MERGE_PARALLEL_EDGES) input->ReorderEdges(true);
		input->ReorderNodesAndEdges();
		C = N = M = 0;
	}
	else
	{
		if (reduction_flags & REDUCTION_REORDER_EDGES)
		{
			input->ReorderEdges((reduction_flags & REDUCTION_MERGE_PARALLEL_EDGES) ? true : false);
		}
		N = -1;
	}
}

GomoryHuReductions::~GomoryHuReductions()
{
}

TmpGraph<GH_REAL>* GomoryHuReductions::GetNextSubproblem(int*& _parents, GH_REAL*& _cut_costs)
{
	if (input->component_num == 0)
	{
		if (N < 0) { N = 0; _parents = parents; _cut_costs = cut_costs; return input; }
		else       return NULL;
	}
	else
	{
		int i;
		if (C > 0)
		{
			if (C != 1)
			{
				for (i=0; i<I.n; i++)
				{
					if (parents[i] >= 0) parents[i] += N;
					else { parents[i] = 0; cut_costs[i] = 0; }
				}
				for (i=0; i<2*I.m; i++) I.edges[i] += N;
				parents -= N;
				cut_costs -= N;
			}

			N += I.n;
			M += I.m;
		}
		if (C < input->component_num)
		{
			I.edges = input->edges + 2*M;
			I.weights = input->weights + M;
			I.n = input->component_sizes[2*C];
			I.m = input->component_sizes[2*C+1];
			if (N != 0) for (i=0; i<2*I.m; i++) I.edges[i] -= N;
			if (input->marks) I.marks = input->marks + N;
			parents += N;
			cut_costs += N;
			_parents = parents;
			_cut_costs = cut_costs;
			C ++;
			return &I;

		}
		int n = input->n;
		int* permutation = input->permutation;
		int* permutation_rev = input->permutation_rev;
		char* buf = new char[n*((sizeof(int) > sizeof(GH_REAL)) ? sizeof(int) : sizeof(GH_REAL))];
		int* _parents = (int*)buf;
		for (i=0; i<n; i++) _parents[i] = (parents[permutation_rev[i]] >= 0) ? permutation[parents[permutation_rev[i]]] : -1;
		memcpy(parents, _parents, n*sizeof(int));
		GH_REAL* _cut_costs = (GH_REAL*)buf;
		for (i=0; i<n; i++) _cut_costs[i] = cut_costs[permutation_rev[i]];
		memcpy(cut_costs, _cut_costs, n*sizeof(GH_REAL));
		delete [] buf;
		I.edges = NULL;
		I.marks = NULL;
		return NULL;
	}
}


#endif