#define _CRT_SECURE_NO_WARNINGS

/*
#########################################################
#                                                       #
#  IBFSGraph -  Software for solving                    #
#               Maximum s-t Flow / Minimum s-t Cut      #
#               using the IBFS algorithm                #
#                                                       #
#  http://www.cs.tau.ac.il/~sagihed/ibfs/               #
#                                                       #
#  Haim Kaplan (haimk@cs.tau.ac.il)                     #
#  Sagi Hed (sagihed@post.tau.ac.il)                    #
#                                                       #
#########################################################

This software implements the IBFS (Incremental Breadth First Search) maximum flow algorithm from
	"Faster and More Dynamic Maximum Flow
	by Incremental Breadth-First Search"
	Andrew V. Goldberg, Sagi Hed, Haim Kaplan, Pushmeet Kohli,
	Robert E. Tarjan, and Renato F. Werneck
	In Proceedings of the 23rd European conference on Algorithms, ESA'15
	2015
and from
	"Maximum flows by incremental breadth-first search"
	Andrew V. Goldberg, Sagi Hed, Haim Kaplan, Robert E. Tarjan, and Renato F. Werneck.
	In Proceedings of the 19th European conference on Algorithms, ESA'11, pages 457-468.
	ISBN 978-3-642-23718-8
	2011

Copyright Haim Kaplan (haimk@cs.tau.ac.il) and Sagi Hed (sagihed@post.tau.ac.il)

###########
# LICENSE #
###########
This software can be used for research purposes only.
If you use this software for research purposes, you should cite the aforementioned papers
in any resulting publication and appropriately credit it.

If you require another license, please contact the above.

*/


#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "ibfs.h"


#ifdef GRAPH_NAMESPACE
namespace GRAPH_NAMESPACE {
#endif


#define REMOVE_SIBLING(x, tmp) \
	{ (tmp) = (x)->parent->head->firstSon; \
	if ((tmp) == (x)) { \
		(x)->parent->head->firstSon = (x)->nextPtr; \
	} else { \
		for (; (tmp)->nextPtr != (x); (tmp) = (tmp)->nextPtr); \
		(tmp)->nextPtr = (x)->nextPtr; \
	} }

#define ADD_SIBLING(x, parentNode) \
	{ (x)->nextPtr = (parentNode)->firstSon; \
	(parentNode)->firstSon = (x); \
	}

IBFSGraph::IBFSGraph(int node_num, int edge_num_max, Buffer* buf, ReusableBuffer* common_rbuf, ReusableBuffer* rbuf)
	: GraphBase(node_num, edge_num_max, buf, common_rbuf, rbuf), prNodeBuckets(orphan3PassBuckets)
{
	incList = NULL;
	incLen = incIteration = 0;
	uniqOrphansS = uniqOrphansT = 0;
	augTimestamp = 0;
	verbose = IBTEST;
	topLevelS = topLevelT = 0;
	flow = 0;
	ptrs = NULL;
	testFlow = 0;
	testExcess = 0;

	/////////////////////////////////////////////////////////////////////

	// allocate nodes
	//	if (verbose) {
	//		fprintf(stdout, "c allocating nodes... \t [%lu MB]\n", (unsigned long)sizeof(Node)*(unsigned long)(numNodes+1) / (1<<20));
	//		fflush(stdout);
	//	}
	active0.init(rbuf+STLISTS_RBUF_NUM);
	activeS1.init(rbuf+STLISTS_RBUF_NUM+1);
	activeT1.init(rbuf+STLISTS_RBUF_NUM+2);
	if (IB_EXCESSES) {
		Node** ptrs = (Node**) buf->Alloc(3*node_num*sizeof(Node*));
		excessBuckets.init(nodes, ptrs, node_num);
	}
	orphan3PassBuckets.init(nodes, node_num);
	orphanBuckets.init(nodes, node_num);

	// init members
	flow = 0;

	if (verbose) {
		fprintf(stdout, "c sizeof(ptr) = %d bytes\n", (int)sizeof(Node*));
		fprintf(stdout, "c sizeof(node) = %d bytes\n", (int)sizeof(Node));
		fprintf(stdout, "c sizeof(arc) = %d bytes\n", (int)sizeof(Arc));
	}
}


IBFSGraph::~IBFSGraph()
{
	orphanBuckets.free();
	orphan3PassBuckets.free();
	excessBuckets.free();
}

void IBFSGraph::initGraph()
{
	init_graph();
	int _x;
	FOR_ALL_NODES(_x)
	{
		Node* x = nodes + _x;
		if (x->excess == 0) {
			x->label = 0;
			continue;
		}
		if (x->excess > 0) {
			x->label = 1;
			activeS1.add(x);
		} else {
			x->label = -1;
			activeT1.add(x);
		}
	}

	topLevelS = topLevelT = 1;
}






// @ret: minimum orphan level
template<bool sTree> int IBFSGraph::augmentPath(Node *x, captype push)
{
	Node *y;
	Arc *a;
	int orphanMinLevel = (sTree ? topLevelS : topLevelT) + 1;

	augTimestamp++;
	for (; ; x=a->head)
	{
		stats.incPushes();
		if (x->excess) break;
		a = x->parent;
		if (sTree) {
			PUSH_FLOW(a, -push);
		} else {
			PUSH_FLOW(a, push);
		}

		// saturated?
		if (sTree ? (!IS_REV_NONSATURATED(a)) : (!IS_NONSATURATED(a)))
		{
			REMOVE_SIBLING(x,y);
			orphanMinLevel = (sTree ? x->label : -x->label);
			orphanBuckets.add<sTree>(x);
		}
	}
	UPDATE_EXCESS(x, (sTree ? -push : push));
	if (x->excess == 0) {
		orphanMinLevel = (sTree ? x->label : -x->label);
		orphanBuckets.add<sTree>(x);
	}
	flow += push;

	return orphanMinLevel;
}


// @ret: minimum level in which created an orphan
template<bool sTree> int IBFSGraph::augmentExcess(Node *x, captype push)
{
	Node *y;
	Arc *a;
	int orphanMinLevel = (sTree ? topLevelS : topLevelT)+1;
	augTimestamp++;

	// start of loop
	//----------------
	// x 		  the current node along the path
	// a		  arc incoming into x
	// push 	  the amount of flow coming into x
	// a->resCap  updated with incoming flow already
	// x->excess  not updated with incoming flow yet
	//
	// end of loop
	//-----------------
	// x 		  the current node along the path
	// a		  arc outgoing from x
	// push 	  the amount of flow coming out of x
	// a->resCap  updated with outgoing flow already
	// x->excess  updated with incoming flow already
	while (sTree ? (x->excess <= 0) : (x->excess >= 0))
	{
		testNode(x);
		stats.incPushes();
		a = x->parent;

		// update excess and find next flow
		if ((sTree ? (GET_RCAP(REV_ARC(a))) : (GET_RCAP(a))) < (sTree ? (push-x->excess) : (x->excess+push))) {
			// some excess remains, node is an orphan
			UPDATE_EXCESS(x, (sTree ? (GET_RCAP(REV_ARC(a)) - push) : (push-GET_RCAP(a))));
			push = (sTree ? GET_RCAP(REV_ARC(a)) : GET_RCAP(a));
		} else {
			// all excess is pushed out, node may or may not be an orphan
			push += (sTree ? -(x->excess) : x->excess);
			SET_EXCESS(x, 0);
		}

		// push flow
		// note: push != 0
		if (sTree) {
			PUSH_FLOW(a, -push);
		} else {
			PUSH_FLOW(a, push);
		}

		// saturated?
		if (sTree ? (!IS_REV_NONSATURATED(a)) : (!IS_NONSATURATED(a)))
		{
			REMOVE_SIBLING(x,y);
			orphanMinLevel = (sTree ? x->label : -x->label);
			orphanBuckets.add<sTree>(x);
			if (x->excess) excessBuckets.incMaxBucket(sTree ? x->label : -x->label);
		}

		// advance
		// a precondition determines that the first node on the path is not in excess buckets
		// so only the next nodes may need to be removed from there
		x = a->head;
		if (sTree ? (x->excess < 0) : (x->excess > 0)) excessBuckets.remove<sTree>(x);
	}

	// update the excess at the root
	if (push <= (sTree ? (x->excess) : -(x->excess))) flow += push;
	else flow += (sTree ? (x->excess) : -(x->excess));
	UPDATE_EXCESS(x, (sTree ? (-push) : push));
	if (sTree ? (x->excess <= 0) : (x->excess >= 0)) {
		orphanMinLevel = (sTree ? x->label : -x->label);
		orphanBuckets.add<sTree>(x);
		if (x->excess) excessBuckets.incMaxBucket(sTree ? x->label : -x->label);
	}

	return orphanMinLevel;
}


template<bool sTree> void IBFSGraph::augmentExcesses()
{
	Node *x;
	int minOrphanLevel;
	int adoptedUpToLevel = excessBuckets.maxBucket;

	if (!excessBuckets.empty())
	for (; excessBuckets.maxBucket != (excessBuckets.minBucket-1); excessBuckets.maxBucket--)
	while ((x=excessBuckets.popFront(excessBuckets.maxBucket)) != NULL)
	{
		minOrphanLevel = augmentExcess<sTree>(x, 0);
		// if we did not create new orphans
		if (adoptedUpToLevel < minOrphanLevel) minOrphanLevel = adoptedUpToLevel;
		adoption<sTree>(minOrphanLevel, false);
		adoptedUpToLevel = excessBuckets.maxBucket;
	}
	excessBuckets.reset();
	if (orphanBuckets.maxBucket != 0) adoption<sTree>(adoptedUpToLevel+1, true);
	// free 3pass orphans
	while ((x=excessBuckets.popFront(0)) != NULL) orphanFree<sTree>(x);
}


void IBFSGraph::augment(Arc *bridge)
{
	Node *x, *y;
	Arc *a;
	captype bottleneck, bottleneckT, bottleneckS;
	int minOrphanLevel;
	bool forceBottleneck;
	stats.incAugs();

	// must compute forceBottleneck once, so that it is constant throughout this method
	forceBottleneck = (IB_EXCESSES ? false : true);
	if (IB_BOTTLENECK_ORIG && IB_EXCESSES)
	{
		// limit by end nodes excess
		bottleneck = GET_RCAP(bridge);
		if (bridge->head->excess != 0 && -(bridge->head->excess) < bottleneck) {
			bottleneck = -(bridge->head->excess);
		}
		if (REV_ARC(bridge)->head->excess != 0 && REV_ARC(bridge)->head->excess < bottleneck) {
			bottleneck = REV_ARC(bridge)->head->excess;
		}
	}
	else
	{
		bottleneck = bottleneckS = GET_RCAP(bridge);
		if (bottleneck != 1) {
			for (x=REV_ARC(bridge)->head; ; x=a->head)
			{
				if (x->excess) break;
				a = x->parent;
				if (bottleneckS > GET_RCAP(REV_ARC(a))) {
					bottleneckS = GET_RCAP(REV_ARC(a));
				}
			}
			if (bottleneckS > x->excess) {
				bottleneckS = x->excess;
			}
			if (IB_EXCESSES && x->label != 1) forceBottleneck = true;
			if (x == REV_ARC(bridge)->head) bottleneck = bottleneckS;
		}

		if (bottleneck != 1) {
			bottleneckT = GET_RCAP(bridge);
			for (x=bridge->head; ; x=a->head)
			{
				if (x->excess) break;
				a = x->parent;
				if (bottleneckT > GET_RCAP(a)) {
					bottleneckT = GET_RCAP(a);
				}
			}
			if (bottleneckT > (-x->excess)) {
				bottleneckT = (-x->excess);
			}
			if (IB_EXCESSES && x->label != -1) forceBottleneck = true;
			if (x == bridge->head && bottleneck > bottleneckT) bottleneck = bottleneckT;

			if (forceBottleneck) {
				if (bottleneckS < bottleneckT) bottleneck = bottleneckS;
				else bottleneck = bottleneckT;
			}
		}
	}

	// stats
	if (IBSTATS) {
		int augLen = (-(bridge->head->label)-1 + REV_ARC(bridge)->head->label-1 + 1);
		stats.addAugLen(augLen);
	}

	// augment connecting arc
	PUSH_FLOW(bridge, bottleneck);
	stats.incPushes();
	flow -= bottleneck;

	// augment T
	x = bridge->head;
	if (!IB_EXCESSES || bottleneck == 1 || forceBottleneck) {
		minOrphanLevel = augmentPath<false>(x, bottleneck);
		adoption<false>(minOrphanLevel, true);
	} else if (IB_ADOPTION_PR && !x->excess) {
		UPDATE_EXCESS(x, bottleneck);
		excessBuckets.add<false>(x);
		REMOVE_SIBLING(x,y);
		augmentExcessesDischarge<false>();
	} else {
		minOrphanLevel = augmentExcess<false>(x, bottleneck);
		adoption<false>(minOrphanLevel, false);
		augmentExcesses<false>();
	}

	// augment S
	x = REV_ARC(bridge)->head;
	if (!IB_EXCESSES || bottleneck == 1 || forceBottleneck) {
		minOrphanLevel = augmentPath<true>(x, bottleneck);
		adoption<true>(minOrphanLevel, true);
	} else if (IB_ADOPTION_PR && !x->excess) {
		UPDATE_EXCESS(x, - bottleneck);
		excessBuckets.add<true>(x);
		REMOVE_SIBLING(x,y);
		augmentExcessesDischarge<true>();
	} else {
		minOrphanLevel = augmentExcess<true>(x, bottleneck);
		adoption<true>(minOrphanLevel, false);
		augmentExcesses<true>();
	}
}




template<bool sTree> void IBFSGraph::adoption(int fromLevel, bool toTop)
{
	Node *x, *y, *z;
	register Arc *a;
	int threePassLevel;
	int minLabel, numOrphans, numOrphansUniq;
	int level;

	threePassLevel=0;
	numOrphans=0;
	numOrphansUniq=0;
	for (level = fromLevel;
		level <= orphanBuckets.maxBucket &&
		(!IB_EXCESSES || toTop || threePassLevel || level <= excessBuckets.maxBucket);
		level++)
	while ((x=orphanBuckets.popFront(level)) != NULL)
	{
		testNode(x);
		stats.incOrphans();
		numOrphans++;
		if (x->lastAugTimestamp != augTimestamp) {
			x->lastAugTimestamp = augTimestamp;
			if (sTree) uniqOrphansS++;
			else uniqOrphansT++;
			numOrphansUniq++;
		}
		if (IB_HYBRID_ADOPTION && threePassLevel == 0 && numOrphans >= 3*numOrphansUniq) {
			// switch to 3pass
			threePassLevel = 1;
		}

		//
		// check for same level connection
		//
		if (x->isParentCurr) {
			a = x->parent;
		} else {
			a = FIRST_ARC(x);
			x->isParentCurr = 1;
		}
		x->parent = NULL;
		if (x->label != (sTree ? 1 : -1))
		{
			minLabel = x->label - (sTree ? 1 : -1);
			FOR_REMAINING_ARCS(x, a)
			{
				stats.incOrphanArcs1();
				y = a->head;
				if ((sTree ? IS_REV_NONSATURATED(a) : IS_NONSATURATED(a)) && y->label == minLabel)
				{
					x->parent = a;
					ADD_SIBLING(x,y);
					break;
				}
			}
		}
		if (x->parent != NULL) {
			if (IB_EXCESSES && x->excess) excessBuckets.add<sTree>(x);
			continue;
		}

		//
		// on the top level there is no need to relabel
		//
		if (x->label == (sTree ? topLevelS : -topLevelT)) {
			orphanFree<sTree>(x);
			continue;
		}

		//
		// give up on same level - relabel it!
		// (1) create orphan sons
		//
		for (y=x->firstSon; y != NULL; y=z)
		{
			stats.incOrphanArcs3();
			z=y->nextPtr;
			if (IB_EXCESSES && y->excess) excessBuckets.remove<sTree>(y);
			orphanBuckets.add<sTree>(y);
		}
		x->firstSon = NULL;

		//
		// (2) 3pass relabeling: move to buckets structure
		//
		if (threePassLevel) {
			x->label += (sTree ? 1 : -1);
			orphan3PassBuckets.add<sTree>(x);
			if (threePassLevel == 1) {
				threePassLevel = level+1;
			}
			continue;
		}

		//
		// (2) relabel: find the lowest level parent
		//
		minLabel = (sTree ? topLevelS : -topLevelT);
		if (x->label != minLabel)
		FOR_ALL_ARCS(x, a)
		{
			stats.incOrphanArcs2();
			y = a->head;
			if ((sTree ? IS_REV_NONSATURATED(a) : IS_NONSATURATED(a)) &&
				// y->label != 0 ---> holds implicitly
				(sTree ? (y->label > 0) : (y->label < 0)) &&
				(sTree ? (y->label < minLabel) : (y->label > minLabel)))
			{
				minLabel = y->label;
				x->parent = a;
				if (minLabel == x->label) break;
			}
		}

		//
		// (3) relabel onto new parent
		//
		if (x->parent != NULL) {
			x->label = minLabel + (sTree ? 1 : -1);
			ADD_SIBLING(x, x->parent->head);
			// add to active list of the next growth phase
			if (sTree) {
				if (x->label == topLevelS) activeS1.add(x);
			} else {
				if (x->label == -topLevelT) activeT1.add(x);
			}
			if (IB_EXCESSES && x->excess) excessBuckets.add<sTree>(x);
		} else {
			orphanFree<sTree>(x);
		}
	}
	if (level > orphanBuckets.maxBucket) orphanBuckets.maxBucket=0;

	if (threePassLevel) {
		adoption3Pass<sTree>(threePassLevel);
	}
}

template <bool sTree> void IBFSGraph::adoption3Pass(int minBucket)
{
	Arc *a;
	Node *x, *y;
	int minLabel, destLabel;

	for (int level=minBucket; level <= orphan3PassBuckets.maxBucket; level++)
	while ((x = orphan3PassBuckets.popFront(level)) != NULL)
	{
		testNode(x);

		// pass 2: find lowest level parent
		if (x->parent == NULL) {
			minLabel = (sTree ? topLevelS : -topLevelT);
			destLabel = x->label - (sTree ? 1 : -1);
			FOR_ALL_ARCS(x, a)
			{
				stats.incOrphanArcs3();
				y = a->head;
				if ((sTree ? IS_REV_NONSATURATED(a) : IS_NONSATURATED(a)) &&
					((sTree ? (y->excess > 0) : (y->excess < 0)) || y->parent != NULL) &&
					(sTree ? (y->label > 0) : (y->label < 0)) &&
					(sTree ? (y->label < minLabel) : (y->label > minLabel)))
				{
					x->parent = a;
					if ((minLabel = y->label) == destLabel) break;
				}
			}
			if (x->parent == NULL) {
				x->label = 0;
				if (IB_EXCESSES && x->excess) excessBuckets.add<sTree>(x);
				continue;
			}
			x->label = minLabel + (sTree ? 1 : -1);
			if (x->label != (sTree ? level : -level)) {
				orphan3PassBuckets.add<sTree>(x);
				continue;
			}
		}

		// pass 3: lower potential sons and/or find first parent
		if (x->label != (sTree ? topLevelS : -topLevelT))
		{
			minLabel = x->label + (sTree ? 1 : -1);
			FOR_ALL_ARCS(x, a)
			{
				stats.incOrphanArcs3();
				y = a->head;

				// lower potential sons
				if ((sTree ? IS_NONSATURATED(a) : IS_REV_NONSATURATED(a)) &&
					(y->label == 0 ||
					(sTree ? (minLabel < y->label) : (minLabel > y->label))))
				{
					if (y->label != 0) orphan3PassBuckets.remove<sTree>(y);
					else if (IB_EXCESSES && y->excess) excessBuckets.remove<sTree>(y);
					y->label = minLabel;
					y->parent = REV_ARC(a);
					orphan3PassBuckets.add<sTree>(y);
				}
			}
		}

		// relabel onto new parent
		ADD_SIBLING(x, x->parent->head);
		x->isParentCurr = 0;
		if (IB_EXCESSES && x->excess) excessBuckets.add<sTree>(x);

		// add to active list of the next growth phase
		if (sTree) {
			if (x->label == topLevelS) activeS1.add(x);
		} else {
			if (x->label == -topLevelT) activeT1.add(x);
		}
	}

	orphan3PassBuckets.maxBucket = 0;
}


template<bool dirS> void IBFSGraph::growth()
{
	Node *x, *y;
	Arc *a;

	int _active;
	for (_active=0; _active<active0.len; _active++)
	{
		// get active node
		x = active0.list[_active];
		testNode(x);

		// node no longer at level
		if (x->label != (dirS ? (topLevelS-1): -(topLevelT-1))) {
			continue;
		}

		// grow or augment
		if (dirS) stats.incGrowthS();
		else stats.incGrowthT();
		Arc* a_next;
		for (a=FIRST_ARC(x); IS_VALID_ARC(x, a); a=a_next)
		{
			a_next = NEXT_ARC(a);
			stats.incGrowthArcs();
			if (dirS ? (!IS_NONSATURATED(a)) : (!IS_REV_NONSATURATED(a))) continue;
			y = a->head;
			if (y->label == 0)
			{
				// grow node x (attach y)
				testNode(y);
				y->isParentCurr = 0;
				y->label = x->label + (dirS ? 1 : -1);
				y->parent = REV_ARC(a);
				ADD_SIBLING(y, x);
				if (dirS) activeS1.add(y);
				else activeT1.add(y);
			}
			else if (dirS ? (y->label < 0) : (y->label > 0))
			{
				// augment
				augment(dirS ? a : REV_ARC(a));
				if (x->label != (dirS ? (topLevelS-1) : -(topLevelT-1))) {
					break;
				}
				if (dirS ? IS_NONSATURATED(a) : IS_REV_NONSATURATED(a)) a_next = a;
			}
		}
	}
	active0.clear();
}

template<bool sTree> void IBFSGraph::augmentIncrements()
{
	Node *x, *y;
	Node **end = incList+incLen;
	int minOrphanLevel = 1<<30;

	for (Node **inc=incList; inc != end; inc++)
	{
		x = (*inc);
#ifdef GRAPH_CONTRACTION
		if (x->is_contracted) continue;
#endif
		testNode(x);
		if (!x->isIncremental || (sTree ? (x->label < 0) : (x->label > 0))) continue;
		x->isIncremental = 0;
		if (x->label == 0)
		{
			//**** new root from outside the tree
			if (!x->excess) continue;
			x->isParentCurr = 0;
			if (x->excess > 0) {
				x->label = topLevelS;
				activeS1.add(x);
			} else if (x->excess < 0) {
				x->label = -topLevelT;
				activeT1.add(x);
			}
		}
		else if ((sTree ? (x->excess <= 0) : (x->excess >= 0)) &&
				(!x->parent || !(sTree ? IS_REV_NONSATURATED(x->parent) : IS_NONSATURATED(x->parent))))
		{
			//**** new orphan
			if (x->parent) REMOVE_SIBLING(x,y);
			if ((sTree ? x->label : -x->label) < minOrphanLevel) {
				minOrphanLevel = (sTree ? x->label : -x->label);
			}
			orphanBuckets.add<sTree>(x);
			if (x->excess) excessBuckets.incMaxBucket(sTree ? x->label : -x->label);
		}
		else if (sTree ? (x->excess < 0) : (x->excess > 0))
		{
			//**** new deficit/excess to empty
			excessBuckets.add<sTree>(x);
		}
		else if (x->excess && x->parent)
		{
			//**** new root
			REMOVE_SIBLING(x,y);
			x->parent = NULL;
			x->isParentCurr = 0;
		}
	}
	if (orphanBuckets.maxBucket != 0) adoption<sTree>(minOrphanLevel, false);
	if (IB_ADOPTION_PR) augmentExcessesDischarge<sTree>();
	else augmentExcesses<sTree>();
}


IBFSGraph::flowtype IBFSGraph::computeMaxFlow()
{
	return computeMaxFlow(true, false);
}

IBFSGraph::flowtype IBFSGraph::computeMaxFlow(bool allowIncrements)
{
	return computeMaxFlow(true, allowIncrements);
}

IBFSGraph::flowtype IBFSGraph::computeMaxFlow(bool initialDirS, bool allowIncrements)
{
	// incremental?
	if (incIteration >= 1 && incList != NULL) {
		augmentIncrements<true>();
		augmentIncrements<false>();
		incList = NULL;
	}

	// test
	if (IBTEST) {
		testFlow = flow;
		int _x;
		FOR_ALL_NODES(_x)
		{
			Node* x = nodes + _x;
			if (x->excess > 0) testExcess += x->excess;
		}
	}

	//
	// IBFS
	//
	bool dirS = initialDirS;
	while (true)
	{
		// BFS level
		if (dirS) {
			ActiveList::swapLists(&active0, &activeS1);
			topLevelS++;
		} else {
			ActiveList::swapLists(&active0, &activeT1);
			topLevelT++;
		}
		orphanBuckets.allocate((topLevelS > topLevelT) ? topLevelS : topLevelT);
		orphan3PassBuckets.allocate((topLevelS > topLevelT) ? topLevelS : topLevelT);
		if (IB_EXCESSES) excessBuckets.allocate((topLevelS > topLevelT) ? topLevelS : topLevelT);
		if (dirS) growth<true>();
		else growth<false>();
		if (IBTEST) {
			testTree();
			fprintf(stdout, "dirS=%d aug=%d   S %d / T %d   flow=%d\n",
					dirS, augTimestamp, uniqOrphansS, uniqOrphansT, (int)flow);
			fflush(stdout);
		}

		// switch to next level
		if (!allowIncrements && (activeS1.len == 0 || activeT1.len == 0)) break;
		if (activeS1.len == 0 && activeT1.len == 0) break;
		if (activeT1.len == 0) dirS=true;
		else if (activeS1.len == 0) dirS=false;
		else if (!IB_ALTERNATE_SMART && dirS) dirS = false;
		else if (IB_ALTERNATE_SMART && uniqOrphansT == uniqOrphansS && dirS) dirS=false;
		else if (IB_ALTERNATE_SMART && uniqOrphansT < uniqOrphansS) dirS=false;
		else dirS=true;
	}

	incIteration++;
	return flow;
}




///////////////////////////////////////////////////
// experimental push relabel orphan processing
///////////////////////////////////////////////////
template<bool sTree> void IBFSGraph::augmentExcessesDischarge()
{
	Node *x;
	if (!excessBuckets.empty())
	for (; excessBuckets.maxBucket != (excessBuckets.minBucket-1); excessBuckets.maxBucket--)
	while ((x=excessBuckets.popFront(excessBuckets.maxBucket)) != NULL) {
		augmentDischarge<sTree>(x);
	}
	excessBuckets.reset();
	while ((x=excessBuckets.popFront(0)) != NULL) {
		x->isIncremental = 0;
		orphanBuckets.add<sTree>(x);
		// TODO: add orphan min level optimization here
	}
	augTimestamp++;
	adoption<sTree>(1, true);
}

// @pre: !x->isIncremental && x not in excessBuckets[0] && x not in x->parent sons list
template<bool sTree> void IBFSGraph::augmentDischarge(Node *x)
{
	Node *y, *z;
	int minLabel;
	captype push;
	Arc *a;
	int startLabel = x->label;
	testNode(x);

	// loop
	while (true)
	{
		// push
		if (x->isParentCurr) {
			a = x->parent;
		} else {
			a = FIRST_ARC(x);
			x->isParentCurr = 1;
		}
		if (x->label != (sTree ? 1 : -1) && a != NULL)
		{
			minLabel = x->label + (sTree ? -1 : 1);
			FOR_REMAINING_ARCS(x, a)
			{
				// check admissible
				y = a->head;
				if ((sTree ? (!IS_REV_NONSATURATED(a)) : (!IS_NONSATURATED(a))) || y->label != minLabel) {
					continue;
				}

				// push admissible
				push = (sTree ? GET_RCAP(REV_ARC(a)) : GET_RCAP(a));
				if (push > (sTree ? (-x->excess) : (x->excess))) {
					push = (sTree ? (-x->excess) : (x->excess));
				}
				UPDATE_EXCESS(x, (sTree ? push : (-push)));
				if (sTree) {
					PUSH_FLOW(a, -push);
				} else {
					PUSH_FLOW(a, push);
				}

				// add excess
				if (sTree && y->excess > 0) {
					if (y->excess >= push) flow += push;
					else flow += y->excess;
				} else if (!sTree && y->excess < 0) {
					if (-y->excess >= push) flow += push;
					else flow -= y->excess;
				}
				UPDATE_EXCESS(y, (sTree ? (-push) : push));
				if (y->excess == 0 /* implicit && !y->isIncremental && y has no parent */) {
					y->label = 0;
					excessBuckets.add<sTree>(y);
					y->label = minLabel;
					y->isIncremental = 1;
				} else if (sTree ? (y->excess < 0 && y->excess >= -push) : (y->excess > 0 && y->excess <= push)) {
					if (y->isIncremental) {
						y->label = 0;
						excessBuckets.remove<sTree>(y);
						y->label = minLabel;
						y->isIncremental = 0;
					} else if (y->parent != NULL) {
						REMOVE_SIBLING(y,z);
					}
					excessBuckets.add<sTree>(y);
				}
				if (x->excess == 0) {
					x->parent = a;
					if (!(sTree ? IS_REV_NONSATURATED(a) : IS_NONSATURATED(a))) {
						x->label = 0;
						excessBuckets.add<sTree>(x);
						x->label = minLabel + (sTree ? 1 : -1);
						x->isIncremental = 1;
					}
					break;
				}
			}
		}
		if (x->excess == 0) break;

		// make sons orphans
		minLabel = x->label + (sTree ? 1 : -1);
		for (y=x->firstSon; y != NULL; y=z)
		{
			stats.incOrphanArcs3();
			z=y->nextPtr;
			// implicit !y->isIncremental && !y->excess
			y->label = 0;
			excessBuckets.add<sTree>(y);
			y->label = minLabel;
			y->isIncremental = 1;
		}
		x->firstSon = NULL;

		// relabel
		minLabel = (sTree ? topLevelS : -topLevelT);
		x->parent = NULL;
		FOR_ALL_ARCS(x, a)
		{
			y = a->head;
			if ((sTree ? IS_REV_NONSATURATED(a) : IS_NONSATURATED(a)) &&
				// y->label != 0 ---> holds implicitly
				(sTree ? (y->label > 0) : (y->label < 0)) &&
				(sTree ? (y->label < minLabel) : (y->label > minLabel)))
			{
				minLabel = y->label;
				x->parent = a;
				if (minLabel == x->label) break;
			}
		}
		if (x->parent != NULL) {
			x->label = minLabel + (sTree ? 1 : -1);
		} else {
			orphanFree<sTree>(x);
			break;
		}
	}

	// set new parent
	if (x->parent != NULL && !x->isIncremental) ADD_SIBLING(x, x->parent->head);
	if (sTree) {
		if (startLabel != x->label && x->label == topLevelS) activeS1.add(x);
	} else {
		if (startLabel != x->label && x->label == -topLevelT) activeT1.add(x);
	}
}




///////////////////////////////////////////////////
// testing/debugging
///////////////////////////////////////////////////
void IBFSGraph::testTree()
{
	Node *x, *y;
	Arc *a;
	double totalExcess=0;

	int _x;
	FOR_ALL_NODES(_x)
	{
		x = nodes + _x;
		if (x->label > topLevelS || x->label < -topLevelT) {
			IBDEBUG("ILLEGAL LABEL!");
			testExit();
		}
		if (x->label == 0) {
			if (x->excess) {
				IBDEBUG("EXCESS OUTSIDE!");
				testExit();
			}
			continue;
		}
		bool sTree = (x->label > 0);
		if (sTree ? (x->excess < 0) : (x->excess > 0)) {
			IBDEBUG("EXCESS ON WRONG SIDE!");
			testExit();
		}
		if (sTree && x->excess > 0) totalExcess += x->excess;
		if (!x->excess && x->parent == NULL) {
			IBDEBUG("NO PARENT!");
			testExit();
		}
		if (!x->excess && x->parent->head->label != x->label + (sTree ? -1 : 1)) {
			IBDEBUG("ILLEGAL PARENT!");
			testExit();
		}
		if (x->label == (sTree ? topLevelS : -topLevelT)) {
			int k=0;
			for (; k < (sTree ? activeS1 : activeT1).len; k++) {
				if ((sTree ? activeS1 : activeT1).list[k] == x) break;
			}
			if (k == (sTree ? activeS1 : activeT1).len && incIteration == 1) {
				IBDEBUG("NOT ACTIVE!");
				testExit();
			}
			continue;
		}
		for (y=x->firstSon; y != NULL; y=y->nextPtr) {
			if (y->parent->head != x) {
				IBDEBUG("ILLEGAL SIBLING!");
				testExit();
			}
		}
		FOR_ALL_ARCS(x, a)
		{
			if (x->isParentCurr &&
					(sTree ? IS_REV_NONSATURATED(a) : IS_NONSATURATED(a)) &&
					(sTree ? (a->head->label > 0) : (a->head->label < 0)) &&
					a->head->label == (sTree ? (x->label-1) : (x->label+1)) &&
					a < x->parent) {
				IBDEBUG("ILLEGAL CURRENT ARC!");
				testExit();
			}
			if (!(sTree ? IS_NONSATURATED(a) : IS_REV_NONSATURATED(a))) continue;
			if (a->head->label > topLevelS || a->head->label < -topLevelT) {
					IBDEBUG("ILLEGAL LABEL!");
					testExit();
			}
			if (a->head->label == 0 || (a->head->parent == NULL && a->head->excess == 0)) {
				IBDEBUG("CROSS OUT NODE!");
				testExit();
			}
			if (sTree ? (a->head->label < 0) : (a->head->label > 0)) {
				IBDEBUG("CROSS NODE!");
				testExit();
			}
			if (sTree ? (a->head->label > (x->label+1)) : (a->head->label < (x->label-1))) {
				IBDEBUG("EXTENDED ARC!");
				testExit();
			}
		}
	}

	if ((int)(testExcess - totalExcess) != (flow - testFlow)) {
//		IBDEBUG("ILLEGAL FLOW!");
//		testExit();
	}
}

void IBFSGraph::testPrint()
{
	int *nums = new int[node_num_noncontracted];
	memset(nums, 0, sizeof(int)*node_num_noncontracted);
	int _x;
	FOR_ALL_NODES(_x)
	{
		Node* x = nodes + _x;
		if (x->label >= 0) {
			nums[x->label]++;
		} else {
			nums[node_num_noncontracted+x->label]++;
		}
	}
	fprintf(stdout, "S = ");
	for (int i=1; i<=topLevelS; i++) {
		fprintf(stdout, "%d ", nums[i]);
	}
	fprintf(stdout, "\nT = ");
	for (int i=1; i<=topLevelT; i++) {
		fprintf(stdout, "%d ", nums[node_num_noncontracted-i]);
	}
	delete []nums;
	fprintf(stdout,"\n");
	fflush(stdout);
}




///////////////////////////////////////////////////
// push relabel implementation
///////////////////////////////////////////////////
void IBFSGraph::pushRelabelShelve(int fromLevel)
{
	Node *x = NULL;
	for (int bucket=fromLevel; bucket <= prNodeBuckets.maxBucket; bucket++) {
		if (prNodeBuckets.isEmpty(bucket)) continue;
//		if (x == NULL) prNodeShelves.add(prNodeBuckets.buckets[bucket]);
//		else x->nextPtr = prNodeBuckets.buckets[bucket];
//		for (x=prNodeBuckets.buckets[bucket]; x->nextPtr != NULL; x = x->nextPtr) {
//			x->label = -(x->label-fromLevel);
//		}
//		x->label = -(x->label-fromLevel);
		for (x=prNodeBuckets.buckets[bucket]; x != NULL; x = x->nextPtr) x->label = 0;
	}
	int numLevels = prNodeBuckets.maxBucket - fromLevel + 1;
	memset(prNodeBuckets.buckets + fromLevel, 0, sizeof(Node*)*numLevels);
	memset(excessBuckets.buckets + fromLevel, 0, sizeof(Node*)*numLevels);
	prNodeBuckets.maxBucket = fromLevel-1;
	excessBuckets.maxBucket = fromLevel-1;
}

//template <bool sTree> void IBFSGraph::prUnshelve(int fromLevel)
//{
//	Node *next;
//	for (Node *x = prNodeShelves.pop(); x != NULL; x=next) {
//		next = x->nextPtr;
//		x->label = fromLevel - x->label;
//		prNodeBuckets.add<sTree>(x);
//		if (x->excess) excessBuckets.add<sTree>(x);
//	}
//}

void IBFSGraph::pushRelabel()
{
	return pushRelabelDir<false>();
}

template<bool sTree> void IBFSGraph::pushRelabelDir()
{
	Node *x;
	int level;

	// init
	topLevelS = topLevelT = node_num_noncontracted;
	pushRelabelGlobalUpdate<sTree>();

	// main loop
	int nDischarges = 0;
	for (; excessBuckets.maxBucket >= excessBuckets.minBucket; excessBuckets.maxBucket--)
	while ((x=excessBuckets.popFront(excessBuckets.maxBucket)) != NULL)
	{
		// discharge
		level = excessBuckets.maxBucket; // excessBuckets.maxBucket may change in discharge()
		pushRelabelDischarge<sTree>(x);
		nDischarges++;
		if (prNodeBuckets.maxBucket < level) {
			excessBuckets.allocate(level+2);
			prNodeBuckets.allocate(level+2);
		}

		// global update / gap heuristic
		if (nDischarges % (30*node_num_noncontracted) == 0) pushRelabelGlobalUpdate<sTree>();
		else if (prNodeBuckets.isEmpty(level)) pushRelabelShelve(level+1);
	}
}

template<bool sTree> void IBFSGraph::pushRelabelGlobalUpdate()
{
	Node *x, *y;
	Arc *a;

	memset(prNodeBuckets.buckets, 0, sizeof(Node*)*(prNodeBuckets.allocLevels+1));
	memset(excessBuckets.buckets, 0, sizeof(Node*)*(excessBuckets.allocLevels+1));
	prNodeBuckets.maxBucket = 1;
	excessBuckets.reset();
	int _x;
	FOR_ALL_NODES(_x)
	{
		x = nodes + _x;
		x->parent = NULL;
		x->isParentCurr = 0;
		if ((sTree ? (x->excess > 0) : (x->excess < 0))) {
			x->label = (sTree ? 1 : -1);
			prNodeBuckets.add<sTree>(x);
		} else x->label = 0;
	}
	for (int bucket=1; bucket <= prNodeBuckets.maxBucket; bucket++)
	for (x=prNodeBuckets.buckets[bucket]; x != NULL; x = x->nextPtr) {
		FOR_ALL_ARCS(x, a)
		{
			if (!(sTree ? IS_NONSATURATED(a) : IS_REV_NONSATURATED(a))) continue;
			y = a->head;
			if (y->parent != NULL || (sTree ? (y->excess > 0) : (y->excess < 0))) continue;
			y->label = (sTree ? (bucket+1) : (-bucket-1));
			prNodeBuckets.add<sTree>(y);
			y->parent = REV_ARC(a);
			if (y->excess) excessBuckets.add<sTree>(y);
		}
	}
}

template<bool sTree> void IBFSGraph::pushRelabelDischarge(Node *x)
{
	Node *y;
	int minLabel;
	captype push;
	Arc *a;

	testNode(x);
	prNodeBuckets.remove<sTree>(x);
	while (true)
	{
		// push
		if (x->isParentCurr) {
			a = x->parent;
		} else {
			a = FIRST_ARC(x);
			x->isParentCurr = 1;
		}
		if (x->label != (sTree ? 1 : -1))
		{
			minLabel = x->label - (sTree ? 1 : -1);
			FOR_REMAINING_ARCS(x, a)
			{
				// check admissible
				y = a->head;
				if ((sTree ? (!IS_REV_NONSATURATED(a)) : (!IS_NONSATURATED(a))) || y->label != minLabel) {
					continue;
				}

				// push admissible
				push = (sTree ? GET_RCAP(REV_ARC(a)) : GET_RCAP(a));
				if (push > (sTree ? (-x->excess) : (x->excess))) {
					push = (sTree ? (-x->excess) : (x->excess));
				}
				UPDATE_EXCESS(x, (sTree ? push : (-push)));
				if (sTree) {
					PUSH_FLOW(a, -push);
				} else {
					PUSH_FLOW(a, push);
				}

				// add excess
				if (sTree && y->excess > 0) {
					if (y->excess >= push) flow += push;
					else flow += y->excess;
				} else if (!sTree && y->excess < 0) {
					if (-y->excess >= push) flow += push;
					else flow -= y->excess;
				}
				UPDATE_EXCESS(y, (sTree ? (-push) : push));
				if (sTree ? (y->excess < 0 && y->excess >= -push) : (y->excess > 0 && y->excess <= push)) {
					excessBuckets.add<sTree>(y);
				}
				if (x->excess == 0) {
					x->parent = a;
					break;
				}
			}
		}
		if (x->excess == 0) break;

		// relabel
		minLabel = (sTree ? (node_num_noncontracted-1) : (-node_num_noncontracted+1));
		x->parent = NULL;
		FOR_ALL_ARCS(x, a)
		{
			y = a->head;
			if ((sTree ? IS_REV_NONSATURATED(a) : IS_NONSATURATED(a)) &&
				// y->label != 0 ---> holds implicitly
				(sTree ? (y->label > 0) : (y->label < 0)) &&
				(sTree ? (y->label < minLabel) : (y->label > minLabel)))
			{
				minLabel = y->label;
				x->parent = a;
				if (minLabel == x->label) break;
			}
		}
		if (x->parent != NULL) {
			x->label = minLabel + (sTree ? 1 : -1);
		} else {
			x->label = 0;
			break;
		}
	}
	if (x->label != 0) prNodeBuckets.add<sTree>(x);
}


#ifdef GRAPH_NAMESPACE
}
#endif

